#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <getopt.h>

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <arpa/inet.h>

#include <linux/if.h>
#include <linux/if_tun.h>


/* Create a TUN device and return a file descriptor for it. The device name
 * (e.g. "tun0") is written to the name string. type should be set to IFF_TUN
 * or IFF_TAP, and header is non-zero in case a 4 byte TUN/TAP header should
 * be included in packets.
 *
 * Return a negative integer on failure, or the file descriptor of the TUN/TAP
 * device on success.
 */
int tun_alloc(char *name, int type, int header) {
    struct ifreq ifr;
    int fd;

    memset(&ifr, 0, sizeof(ifr));
    switch(type) {
        case IFF_TUN:
            strcpy(ifr.ifr_name, "tun%d");
            break;
        case IFF_TAP:
            strcpy(ifr.ifr_name, "tap%d");
            break;
        default:
            return -1;
    }

    if(header)
        ifr.ifr_flags = type;
    else
        ifr.ifr_flags = type | IFF_NO_PI;

    if((fd = open("/dev/net/tun", O_RDWR)) < 0) return -1;
    if(ioctl(fd, TUNSETIFF, &ifr) < 0) {
        close(fd);
        return -1;
    }
    strcpy(name, ifr.ifr_name);
    return fd;
}

void help(void) {
    fprintf(stderr,
"tunein --ip4 <IPv4 address> [--ip4mask <netmask>] [--ip6 <IPv6 address>\n"
"       [--ip6route] [--type tun|tap] [--header]\n"
"\n"
"   --ip4       -4      IPv4 address of the created interface.\n"
"   --ip4mask   -m      IPv4 netmask.\n"
"   --ip6       -6      IPv6 address of the created interface.\n"
"   --ip6route  -r      Use the interfaces as your default IPv6 route.\n"
"   --type      -t      Determine the type of the interface. Allowed values\n"
"                       are \"tun\" or \"tap\".\n"
"   --noheader  -d      Don't use any TUN/TAP header. May break IPv6.\n"
"\n"
    );
}

int main(int argc, char **argv) {
    int n, i, c, fd;
    fd_set rdfs;

    int type = IFF_TUN;
    int header = 1;
    char address4[64];
    char address6[256];
    char netmask[64];
    char iface_name[16];
    char buffer[4096];
    int ip6route = 0;

    address4[0] = 0;
    address6[0] = 0;
    strcpy(netmask, "255.255.255.0");

    if(argc == 1) {
        help();
        exit(1);
    }

    for(;;) {
        int i = 0;
        static struct option options[] = {
            { "help",       1,  NULL,   'h' },
            { "type",       1,  NULL,   't' },
            { "noheader",   0,  NULL,   'd' },
            { "ip4mask",    1,  NULL,   'm' },
            { "ip4",        1,  NULL,   '4' },
            { "ip6",        1,  NULL,   '6' },
            { "ip6route",   0,  NULL,   'r' }
        };
        if((c = getopt_long(argc, argv, "t:4:m:6:rdh", options, &i)) == -1)
            break;
        switch(c) {
            case '?':
                help();
                exit(1);
            case 'h':
                help();
                exit(0);
            case 't':
                if(!strcmp(optarg, "tap"))
                    type = IFF_TAP;
                else if(!strcmp(optarg, "tun"))
                    type = IFF_TUN;
                else {
                    fprintf(stderr, "%s: invalid device type.\n", argv[0]);
                    exit(1);
                }
                break;
            case 'd':
                header = 0;
                break;
            case '4':
                strncpy(address4, optarg, sizeof(address4));
                break;
            case 'm':
                strncpy(netmask, optarg, sizeof(netmask));
                break;
            case '6':
                strncpy(address6, optarg, sizeof(address6));
                break;
            case 'r':
                ip6route = 1;
                break;
        }
    }
    if(!*address4 && !*address6) {
        fprintf(stderr, "%s: no IP address specified.\n", argv[0]);
        exit(1);
    }

    if((fd = tun_alloc(iface_name, type, header)) < 0) {
        fprintf(stderr, "%s: can not initialize TUN/TAP device.\n", argv[0]);
        exit(1);
    }

    snprintf(buffer, sizeof(buffer), "/sbin/ifconfig %s ", iface_name);
    if(*address4) {
        strcat(buffer, address4);
        strcat(buffer, " netmask ");
        strcat(buffer, netmask);
    }
    if(*address6) {
        strcat(buffer, " add ");
        strcat(buffer, address6);
    }
    if(system(buffer)) {
        fprintf(stderr, "%s: can not set up interface.\n", argv[0]);
        exit(1);
    }
    if(*address6 && ip6route) {
        system("route -A inet6 del 2000::/3 2>/dev/null");
        snprintf(buffer, sizeof(buffer), "route -A inet6 add 2000::/3 dev %s",
            iface_name);
        if(system(buffer)) {
            fprintf(stderr, "%s: can not set IPv6 route.\n", argv[0]);
            exit(1);
        }
    }

    for(;;) {
        FD_ZERO(&rdfs);
        FD_SET(0, &rdfs);
        FD_SET(fd, &rdfs);
        select(fd+1, &rdfs, NULL, NULL, NULL);
        if(FD_ISSET(0, &rdfs)) {
            n = read(0, buffer, sizeof(buffer));
            if(n <= 0) {
                close(fd);
                exit(0);
            }
            if(write(fd, buffer, n) < 0) {
                close(0);
                exit(0);
            }
        }
        if(FD_ISSET(fd, &rdfs)) {
            n = read(fd, buffer, sizeof(buffer));
            if(n <= 0) {
                close(0);
                exit(0);
            }
            if(write(1, buffer, n) < 0) {
                close(fd);
                exit(0);
            }
        }
    }

    return 0;
}

